require("BoxGeometry")

Level = {}

function Level:New(o)
	o = o or {}   -- create object if user does not provide one
	setmetatable(o, self)
	self.__index = self
	
	local textures = { "Texture/blue.dds", "Texture/orange.dds", "Texture/purple.dds" }
	self.materials = {}
	for i, v in pairs(textures) do
		self.materials[i] = Graphics.Material("Effect/blinn")
		self.materials[i]:SetTexture(
			self.materials[i]:GetParameterByName("ColorTexture"), 
			Graphics.Texture(v))
	end
	
	-- Currently it is needed to manually keep reference of lua created objects. 
	-- Otherwise they will be removed by gc and C++ side cannot access.
	self.geometries = {}
	self.colliders = {}
	
	self:CreateStaticBox(Math.Vector3(-50, -10, -50), Math.Vector3(50, 0, 50))
	
	self:CreateStair(Math.Vector3(0, 0, -15), 8, 10, 0.5, 2)
	
	for x = -3, 3, 2 do
		for z = -8, -4, 2 do
			self:CreateMovableBox(Math.Vector3(x, 0.5, z))
		end
	end
	
	self.platforms = {
		Platform:New { 
			positionA = Math.Vector3(10, 1.5, 0), 
			positionB = Math.Vector3(10, 3.5, 0), 
			material = self.materials[3],
			period = 4
		},

		Platform:New { 
			positionA = Math.Vector3(10, 6.5, 6), 
			positionB = Math.Vector3(10, 4.5, 6), 
			material = self.materials[3],
			period = 4
		},

		Platform:New { 
			positionA = Math.Vector3(10, 0.5, -7.5), 
			positionB = Math.Vector3(10, 0.5, -12.5), 
			extent = Math.Vector3(5, 1, 10),
			material = self.materials[3],
			period = 4
		}
	}
	
	self:CreateMovableBox(Math.Vector3(10, 4.0, 0))
	self:CreateMovableBox(Math.Vector3(10, 1.5, -7.5))
	
	return o
end

function Level:OnBeforeUpdate()
	for i, platform in ipairs(self.platforms) do 
		platform:OnBeforeUpdate()
	end
end

function Level:CreateStaticBox(min, max)
	local position = (min + max) * 0.5
	local extent = max - min
	
	local geometry = CreateBoxGeometry(extent)
	self.geometries[#self.geometries + 1] = geometry

	local world = Engine.Game_Instance():GetWorld()
	local box = world:CreateEntity()
	box:SetName("staticBox")
	box:SetLocalTransform(Math.Matrix44_Translate(position))

	local mesh = Engine.Mesh_Cast(box:CreateComponent(Engine.MeshComponentType))
	mesh:SetGeometry(geometry)
	mesh:SetMaterial(self.materials[1])

	local physical = Engine.Physical_Cast(box:CreateComponent(Engine.PhysicalComponentType))
	physical:GetRigidBody():SetMass(0)
	
	local collider = Physics.BoxCollider()
	collider:SetHalfExtent(extent * 0.5)
	physical:GetRigidBody():SetCollider(collider)
	self.colliders[#self.colliders + 1] = collider
end

function Level:CreateStair(position, level, width, riser, tread)
	for i = 0, level do
		self:CreateStaticBox(
			position + Math.Vector3(-width * 0.5, i * riser, ( - i) * tread), 
			position + Math.Vector3(width * 0.5, (i + 1) * riser, (1 - i) * tread))
	end
end

function Level:CreateMovableBox(position, extent)
	extent = extent or Math.Vector3(1, 1, 1)
	
	local world = Engine.Game_Instance():GetWorld()
	local box = world:CreateEntity()
	box:SetName("movableBox")
	box:SetLocalTransform(Math.Matrix44_Translate(position))

	local geometry = CreateBoxGeometry(extent)
	self.geometries[#self.geometries + 1] = geometry

	local mesh = Engine.Mesh_Cast(box:CreateComponent(Engine.MeshComponentType))
	mesh:SetGeometry(geometry)
	mesh:SetMaterial(self.materials[2])

	local physical = Engine.Physical_Cast(box:CreateComponent(Engine.PhysicalComponentType))
	
	local collider = Physics.BoxCollider()
	collider:SetHalfExtent(extent * 0.5)
	physical:GetRigidBody():SetCollider(collider)	
	self.colliders[#self.colliders + 1] = collider
end

Platform = {}

function Platform:New(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self

	o.extent = o.extent or Math.Vector3(5, 1, 5)
	o.time = o.time or 0
	o.period = o.period or 2
	assert(o.material)
	assert(o.positionA)
	assert(o.positionB)
	
	local world = Engine.Game_Instance():GetWorld()
	o.entity = world:CreateEntity()
	o.entity:SetName("platform")
	o.entity:SetLocalTransform(Math.Matrix44_Translate(o.positionA))

	o.geometry = CreateBoxGeometry(o.extent)

	local mesh = Engine.Mesh_Cast(o.entity:CreateComponent(Engine.MeshComponentType))
	mesh:SetGeometry(o.geometry)
	mesh:SetMaterial(o.material)

	o.physical = Engine.Physical_Cast(o.entity:CreateComponent(Engine.PhysicalComponentType))
	o.physical:GetRigidBody():SetMass(0)
	
	o.collider = Physics.BoxCollider()
	o.collider:SetHalfExtent(o.extent * 0.5)
	o.physical:GetRigidBody():SetCollider(o.collider)	
	
	return o
end

function Platform:OnBeforeUpdate()
	local t = (self.time / self.period) % 1
	local velocity = (self.positionB - self.positionA) * (2 / self.period)
	
	if t < 0.5 then
		t = t * 2
	else
		t = (1 - t) * 2
		velocity = -velocity
	end
	
	local position = Math.Vector3_Lerp(self.positionA, self.positionB, t)
	self.physical:GetRigidBody():SetWorldTransform(Math.Matrix44_Translate(position))
	self.physical:GetRigidBody():SetLinearVelocity(velocity)

	self.time = (self.time + Engine.Game_Instance():GetTimeStep()) % self.period
end